
import os
import json
import trimesh
import numpy as np
from tqdm import tqdm 
from langchain_openai import OpenAIEmbeddings
from sceneprogllm import LLM
from dotenv import load_dotenv
load_dotenv()

FUTURE_PATH_MODELS = os.getenv('FUTURE_PATH_MODELS')
HSSD_PATH_MODELS = os.getenv('HSSD_PATH_MODELS')
FUTURE_PATH_IMAGES = os.getenv('FUTURE_PATH_IMAGES')
HSSD_PATH_IMAGES = os.getenv('HSSD_PATH_IMAGES')

from pathlib import Path
BASE = str(Path(__file__).parent)
embd_location = os.path.join(BASE,'all_embeddings.npz')
metadata_location = os.path.join(BASE,'all_metadata.json')

def get_image_path(assetID):
    if assetID.startswith("future/"):
        assetID = assetID.split("/")[1]
        return os.path.join(FUTURE_PATH_IMAGES, assetID + ".png")
    elif assetID.startswith("hssd/"):
        assetID = assetID.split("/")[1]
        return os.path.join(HSSD_PATH_IMAGES, assetID + ".png")
    else:
        raise ValueError(f"Unknown assetID: {assetID}")
    
class AssetRetrieverFutureHSSD:
    def __init__(self):
        self.embeddings = OpenAIEmbeddings(model="text-embedding-3-large", api_key=os.getenv("OPENAI_API_KEY"))
        data = np.load(embd_location, allow_pickle=True)
        self.all_embeddings = data['all_embeddings']
        self.all_models = data['all_models']

        with open(metadata_location, 'r') as f:
            self.metadata = json.load(f)
        
        self.llm = LLM(
            name="AssetRetrieverFutureHSSD",
            system_desc=f"""
Your task is to help me pick a 3D model by given a few images of candidate models based on the closeness to the query. 
Just responsd with the ID of the model.
""",
        response_format="json",
        json_keys=["name:str"],
        )

    def __call__(self, query, context="None"):
        # Step 1: Embed query and compute similarity
        emb = np.array(self.embeddings.embed_query(query))
        similarity = np.dot(self.all_embeddings, emb)

        # Get top 20 most similar models
        top_indices = np.argsort(similarity)[-20:][::-1]
        top_models = [self.all_models[i] for i in top_indices]
        top_similarities = similarity[top_indices]

        # Step 2: If similarity is too low, reject
        if np.max(top_similarities) < 0.5:
            return "No models found"

        # Normalize similarity to get probabilities (softmax over adjusted scores)
        adjusted_scores = 5 * (top_similarities - 0.4)
        probabilities = np.exp(adjusted_scores) / np.sum(np.exp(adjusted_scores))

        # Step 3: Sample 5 models from the 20, based on their probabilities
        sampled_indices = np.random.choice(len(top_models), size=5, replace=False, p=probabilities)
        sampled_models = [str(top_models[i]) for i in sampled_indices]

        image_paths = [get_image_path(model) for model in sampled_models]
        metadata = ""
        for model in sampled_models:
            metadata += f"""
Model ID: {model}
Description: {self.metadata[model]['description']}
Is there free space on top of the object?: {self.metadata[model]['freetop']}
Typical placement of the object: {self.metadata[model]['placement']}
Can this object be placed on top of or inside another object?: {self.metadata[model]['on_top_or_inside']}
Width of the object in meters: {self.metadata[model]['scale']}
"""
        prompt = f"""
Pick the most relevant model for the query: {query}
Following is some additional context provided by the user to help you make a decision: {context}
Pick from the following models:
{metadata }

Return the ID of the model you think is the best fit.
"""
        response = self.llm(prompt, image_paths=image_paths)
        return response["name"]


from dotenv import load_dotenv
load_dotenv()

class SceneProgAssetRetriever:
    def __init__(self):
       
        import os
        from pathlib import Path
        path = Path(__file__).parent
        
        if not os.path.exists(os.path.join(path, '.env')):
            msg = f"""
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
You need to create a .env file first!
cp .env {os.path.join(path, '.env')}
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
"""
            raise Exception(msg)

        if not os.path.exists(os.path.join(path,'futurehssd/all_metadata.json')) or not os.path.exists(os.path.join(path,'objaverse/compiled.json')):
            msg = f"""
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
You need to download the assets first!

Run the following commands to download the assets:

aws s3 cp s3://{os.getenv("DATASET_METADATA_REPO")}/futurehssd/ {os.path.join(path,'futurehssd/')} --recursive
aws s3 cp s3://{os.getenv("DATASET_METADATA_REPO")}/objaverse/ {os.path.join(path,'objaverse/')} --recursive
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
"""
            raise Exception(msg)
       
        from .futurehssd.futurehssd import AssetRetrieverFutureHSSD 
        from .objaverse.objaverse import AssetRetrieverObjaverse

        self.future_hssd = AssetRetrieverFutureHSSD()
        self.objaverse = AssetRetrieverObjaverse()
    
    def __call__(self, description, context="None"):
        ## first search in Future
        future_hssd_results = self.future_hssd(description, context=context)
        if not future_hssd_results == 'No models found':
            return future_hssd_results
        
        ## then search in OBJAVERSE
        objaverse_results = self.objaverse(description)
        if not objaverse_results == 'No models found':
            return ("OBJAVERSE",objaverse_results)
        
        ## Lastly create the model!
        create_model_results = create_model(description)
        return create_model_results

def create_model(description):
    import os
    import requests
    from scipy.spatial.transform import Rotation as R
    import numpy as np
    import trimesh
    response = requests.post(os.getenv("GENERATE_GLB"), json={"text": description})
    cwd = os.getcwd()
    os.makedirs(f"{cwd}/tmp", exist_ok=True)
    random_name = os.urandom(16).hex()
    path = f"{cwd}/tmp/{random_name}.glb"
    if response.status_code == 200:
        with open(path, "wb") as f:
            f.write(response.content)
    else:
        raise Exception("Failed to generate 3D model")

    mesh = trimesh.load(path, force="mesh")
    rotation = R.from_euler('y', 30, degrees=True)
    rotation = np.hstack((rotation.as_matrix(), np.array([[0], [0], [0]])))
    rotation = np.vstack((rotation, np.array([0, 0, 0, 1])))
    mesh.apply_transform(rotation)
    mesh.export(path)
    return path


import os
import json
import bentoml
import trimesh
import numpy as np
from scipy.spatial.transform import Rotation as R

class AssetRetrieverObjaverse:
    def __init__(self):
        self.img2rot = {
            "0.png": 0,
            "1.png": 30,
            "2.png": 60,
            "3.png": 90,
            "4.png": 120,
            "5.png": 150,
            "6.png": -180,
            "7.png": -150,
            "8.png": -120,
            "9.png": -90,
            "10.png": -60,
            "11.png": -30,
        }
        from pathlib import Path
        BASE = str(Path(__file__).parent)
        self.correction_data  = os.path.join(BASE, 'compiled.json')
        with open(self.correction_data, 'r') as f:
            self.correction_data = json.load(f)
        
        self.OBJAVERSE_PATH = os.getenv('OBJAVERSE_PATH')

    def rotate_vertices_around_point(self, vertices, rot, point):
        x,y,z = point
        vertices -= np.array([[x,y,z]])
        r = R.from_euler('y', rot, degrees=True).as_matrix()
        vertices = (r@vertices.T).T
        vertices += np.array([[x,y,z]])
        return vertices

    def apply_objaverse_fix_init(self, mesh):
        vertices = mesh.vertices
        vertices -= vertices.mean(axis=0)
        vertices = self.rotate_vertices_around_point(vertices, 180, vertices.mean(axis=0))
        mesh.vertices = vertices
        return mesh

    def apply_objaverse_fix(self, mesh, rot):
        vertices = mesh.vertices
        vertices = self.rotate_vertices_around_point(vertices, rot, vertices.mean(axis=0))
        mesh.vertices = vertices
        return mesh

    def get_objaverse_local(self, obj):
        path = os.path.join(self.OBJAVERSE_PATH,f'{obj}.glb')
        mesh = trimesh.load(path, force='mesh', process=False)
        mesh = self.apply_objaverse_fix_init(mesh)
        return mesh
    
    def get_id(self, desc):
        with bentoml.SyncHTTPClient('http://chetak.ucsd.edu:3001') as client:
            result: str = client.retrieve([desc])[0]

        if result == '**':
            return "No models found"
        assetID, bbox = result.split('|')

        return assetID

    def __call__(self, desc):
        id = self.get_id(desc)

        if id == "No models found":
            return None
        
        mesh = self.get_objaverse_local(id)
        if id in self.correction_data:
            mesh = self.apply_objaverse_fix(mesh, self.img2rot[self.correction_data[id]])
        
        return mesh




from setuptools import setup, find_packages
setup(
    name="sceneprogdatasets",  # Package name
    version="0.1.8",            # Version number
    author="Kunal Gupta",
    author_email="k5gupta@ucsd.edu",
    description="A dataset retriever for SceneProg project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/KunalMGupta/sceneprogdatasets.git",  # Update with actual URL
    packages=find_packages(include=["sceneprogdatasets", "sceneprogdatasets.*"]),
    include_package_data=True,  # Ensures compiled.json is included
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",  # Define minimum Python version
    install_requires=[
        'sceneprogllm',
        'scipy',
        'trimesh'
    ],
)


# sceneprogdatasets/__init__.py

from .retriever import *  # Import core functionality
# from .future.future import *  # Import future module
# from .HSSD.hssd import *  # Import HSSD module
# from .objaverse.objaverse import *  # Import objaverse module

__version__ = "0.1.0"